@@ -2,6 +2,7 @@
 APINexus
 Class definition
 """
+import datetime
 import json
 
 import pandas
@@ -62,12 +63,13 @@ class APINexus:
             IP_Maquina (str): url de nexus
             Puerto (int): puerto de la API. 56000 de forma predeterminada
             token (str): token de la API
-            version (str): 1.0 o 2.0 de momento. 1.0 by default
+            version (str): 1.0, 2.0 or 3.0
         """
         self.IP_Maquina = IP_Maquina
         self.Puerto = str(Puerto)
         self.token = token
         self.version = version
+        self._version_number = float(version)
         self.url_NX = "http://" + IP_Maquina + ":" + str(Puerto)
         self.header = {
             "nexustoken": self.token,
@@ -79,18 +81,17 @@ class APINexus:
 
     def log_print(self, message, severity='info'):
         if self.logger is not None:
-            if severity=='info':
+            if severity == 'info':
                 self.logger.info(message)
-            elif severity=='error':
+            elif severity == 'error':
                 self.logger.error(message)
-            elif severity=='debug':
+            elif severity == 'debug':
                 self.logger.debug(message)
             else:
                 raise ValueError('Severity not supported')
         else:
             print(message)
 
-
     def __json_normalize_check(self, response) -> pd.DataFrame:
         """
         Intenta convertir a DataFrame la respuesta en json de la API. Si no es posible, es que ha habido un fallo (la respuesta no sigue el formato establecido)
@@ -123,18 +124,18 @@ class APINexus:
         url_completa = self.url_NX + "/api/Documents"
         return self.__getResponse(url_completa)
 
-    def callGetTagViews(self, uid):
+    def callGetTagViews(self, uid: str):
         """Lectura de variables contenidas en una vista. Como parametros recibe unicamente el uid de la vista que queremos leer ya que el token , la IP y el puerto ya se han definido en la instanciacion de la clase"""
         url = self.url_NX + "/api/Documents/tagviews/" + uid
         return self.__getResponse(url)
 
-    def callGetTagViewsRealTime(self, uid, uids_vbles):
+    def callGetTagViewsRealTime(self, uid: str, uids_vbles: list):
         """Lectura de variables de una vista. Devuelve el valor en tiempo real de las variables establecidas en el array uids, contenidas en la vista uid """
         body = json.dumps(uids_vbles)
         url = self.url_NX + "/api/Documents/tagviews/" + uid + "/realtime"
         return self.__postResponse(url, body)
 
-    def callGetDataviewHistory(self, uid, nexusRequest):
+    def callGetDataviewHistory(self, uid: str, nexusRequest: NexusRequest):
         """Lectura de valores historicos de variables"""
         body = json.dumps(
             nexusRequest, default=lambda o: o.__dict__, sort_keys=True, indent=2
@@ -142,6 +143,15 @@ class APINexus:
         url = self.url_NX + "/api/Documents/tagviews/" + uid + "/historic"
         return self.__postResponse(url, body)
 
+    def callGetDataview_History(self, uid, uids, StartTs, EndTs, dataSource, resolution):
+        """Lectura de valores historicos de variables a partir de los parámetros pasados a la funcion
+        StartTS como timestamp y EndTs como timestamp"""
+        body = json.dumps(
+            {"uids": uids, "startTs": StartTs, "endTs": EndTs, "dataSource": dataSource, "resolution": resolution})
+
+        url = self.url_NX + "/api/Documents/tagviews/" + uid + "/historic"
+        return self.__postResponse(url, body)
+
     def callGetTagsWritable(self):
         """Metodo de consultas de tags escribibles"""
         url = self.url_NX + "/api/Tags/writable"
@@ -152,6 +162,24 @@ class APINexus:
         url = self.url_NX + "/api/Tags"
         return self.__getResponse(url)
 
+    def callGetTags_Attributes(self):
+        """Método de consulta de los atributos de los tags de la instalación"""
+        if self._version_number < 3:
+            raise NotImplementedError('Nexus API version must be greater than 3.0')
+        url = self.url_NX + "/api/Tags/Attributes"
+        return self.__getResponse(url)
+
+    def callGetTagswithAtt(self, bool_att=True):
+        """Método de consulta de tags de la instalación
+        Se incluye el parámetro bool_att que permite mostrar o no los
+        attributos a partir de la versión V3.0
+        """
+        if self._version_number < 3:
+            raise NotImplementedError('Nexus API version must be greater than 3.0')
+        url = self.url_NX + "/api/Tags?IncludeAttributes=" + str(bool_att)
+
+        return self.__getResponse(url)
+
     def callGetTagsRealTime(self, uids_vbles):
         """Devuelve valor RT de los tags de la instalacion definidos en un array UIDS_VBLES"""
         body = json.dumps(uids_vbles)
@@ -159,39 +187,105 @@ class APINexus:
         return self.__postResponse(url, body)
 
     def callGetTagsHistory(self, nexusRequest):
-        """Devuelve valor historico de los tags especificados en la estructura NexusRequest"""
+        """Deprecated. Devuelve valor historico de los tags especificados en la estructura NexusRequest"""
         body = json.dumps(nexusRequest, default=lambda o: o.__dict__, sort_keys=True, indent=2)
         url = self.url_NX + "/api/Tags/historic"
         return self.__postResponse(url, body)
 
+    def callGetTags_History(self, uids, StartTs, EndTs, dataSource, resolution, aggOperation="LAST_VALUE"):
+        """Devuelve valor historico de los tags especificados en los parametros, añade el parámetro aggOperation"""
+        if self._version_number < 3:
+            raise NotImplementedError('Nexus API version must be greater than 3.0')
+        if len(uids) > 100:
+            raise AttributeError('Too many tags. Please set the number of tags less or equal than 100')
+        body = json.dumps(
+            {"uids": uids, "startTs": StartTs, "endTs": EndTs, "dataSource": dataSource, "resolution": resolution,
+             "aggOperation": aggOperation})
+        url = self.url_NX + "/api/Tags/historic"
+        return self.__postResponse(url, body)
+
+    def callGetTags_rawHistory(self, uids, StartTs, EndTs):
+        """Devuelve valor historico de los tags especificados en los parametros devolviendo solo los valores raw"""
+        if self._version_number < 3:
+            raise NotImplementedError('Nexus API version must be greater than 3.0')
+        body = json.dumps({"uids": uids, "startTs": StartTs, "endTs": EndTs})
+        url = self.url_NX + "/api/Tags/rawhistoric"
+
+        return self.__postResponse(url, body)
+
+    def callGetAPITags_Filtered(self, installations: list = None, drivers: list = None, tags: list = None,
+                                attributes: list = None):
+        """Busca tags disponibles en el token filtrando por distintos elementos
+        Args:
+            installations:
+            drivers:
+            tags:
+            attributes (not implemented yet)
+            """
+        if self._version_number < 3:
+            raise NotImplementedError('Nexus API version must be greater than 3.0')
+        # TODO: El metodo original permitirá filtrar por atributos pero como se desconoce como hacerlo y estamos en depuración de momento dejo solo estas tres opciones
+        args = locals()
+        for _, arg in args.items():
+            if arg is None:
+                arg = []
+        body = json.dumps(
+            {"SearchByInstallationName": installations, "SearchByDriverName": drivers, "SeachByTagName": tags})
+        url = self.url_NX + "/api/Tags/filtered"
+
+        return self.__postResponse(url, body)
+
+    def callPostTags_InsertorUpdate(self, NameofTag, nameofattribute1="", valueofattribute1="", nameofattribute2="",
+                                    valueofattribute2="", nameofattribute3="", valueofattribute3=""):
+        """Permite actualizar un tag o crearlo con atributos, se permiten 3 atributos en pack de nombre de atributo + valor de attributo
+        """
+        if self._version_number < 3:
+            raise NotImplementedError('Nexus API version must be greater than 3.0')
+        body = json.dumps([{"Name": NameofTag, "Attributes": [
+            {"attributeName": nameofattribute1,
+             "value": valueofattribute1
+             },
+            {"attributeName": nameofattribute2,
+             "value": valueofattribute2
+             },
+            {"attributeName": nameofattribute3,
+             "value": valueofattribute3
+             }
+        ]
+                            }
+                           ])
+
+        url = self.url_NX + "/api/Tags/InsertOrUpdate"
+        return self.__postResponse(url, body)
+
     # TODO: Eliminar el header personalizado de alarmas cuando la versión de la API sea acumulativa
     @WarningsAndJson
     def callGetAlarms(self):
         """Lectura de alarmas compartidas con el mismo token"""
-        header = self.header.copy()
-        header['nexusapiversion'] = '2.0'
+        if self._version_number < 2:
+            raise NotImplementedError('Version 1.0 does not support alarm methods')
         url_completa = self.url_NX + "/api/Alarms"
-        return requests.get(url_completa, verify=False, headers=header)
+        return requests.get(url_completa, verify=False, headers=self.header)
 
     @WarningsAndJson
     def callGetAlarmByuid(self, uid):
         """Lectura de alarmas compartidas con el mismo token"""
-        header = self.header.copy()
-        header['nexusapiversion'] = '2.0'
+        if self._version_number < 2:
+            raise NotImplementedError('Version 1.0 does not support alarm methods')
         url_completa = self.url_NX + "/api/Alarms/alarm/" + uid
-        return requests.get(url_completa, verify=False, headers=header)
+        return requests.get(url_completa, verify=False, headers=self.header)
 
     # @WarningsAndJson -- no se necesita ya que se requiere comparar el objeto response
     def callPostAckAlarm(self, uid, status: str):
         """Lectura de alarmas compartidas con el mismo token
         status: 'ARE' or 'EXR'
         """
+        if self._version_number < 2:
+            raise NotImplementedError('Version 1.0 does not support alarm methods')
         if status == 'ARE' or status == 'EXR':
-            header = self.header.copy()
-            header['nexusapiversion'] = '2.0'
             url_completa = self.url_NX + "/api/Alarms/alarm/" + uid + "/acknowledge"
             body = "\"" + status + "\""
-            return requests.post(url_completa, verify=False, headers=header, data=body)
+            return requests.post(url_completa, verify=False, headers=self.header, data=body)
         else:
             raise ValueError("Status must be 'ARE' or 'EXR'")
 
@@ -202,40 +296,39 @@ class APINexus:
         args:
             uid: uid de la alarma
         """
-        header = self.header.copy()
-        header['nexusapiversion'] = '2.0'
+        if self._version_number < 2:
+            raise NotImplementedError('Version 1.0 does not support alarm methods')
         url_completa = self.url_NX + "/api/Alarms/alarm/" + uid + "/event"
-        return requests.post(url_completa, verify=False, headers=header, json={'Message': msg})
+        return requests.post(url_completa, verify=False, headers=self.header, json={'Message': msg})
 
-    def callPostTagInsert(self, variable_name):
-        """consulta de tags con permisos de escritura"""
+    def get_writable_tags(self) -> pd.DataFrame:
+        """
+        Get all writable tags from token.
+        Returns:
+             data: dataframe with tag info. If version is v3, will also return attribute info
+        """
         url_get = self.url_NX + "/api/Tags/writable"
         variables = self.__getResponse(url_get)
-        variables_names = list()
-        try:
-            variables_norm = self.__json_normalize_check(variables)
-            variables_names = list(variables_norm.name)
-        except:
-            self.log_print(variables_names)
+        data = self.__json_normalize_check(variables)
+        return data
 
-        if not variable_name in variables_names:
-            self.log_print("Se procede a crear la variable con nombre "
-                  + variable_name
-                  + ". Se devuelve json con uid y nombre de variable creada"
-                  )
-            payload = '["' + variable_name + '"]'
-            self.log_print(payload)
-            url_post = self.url_NX + "/api/Tags/insert"
-            self.log_print(url_post)
-            response = self.__postResponse(url_post, payload)
-            dataReceived = self.__json_normalize_check(response)
-        else:
-            self.log_print(
-                "La variable ya existe, no se creará ninguna variable. Se devuelve el json con el uid de la variable existente"
-            )
-            dataReceived = variables_norm[variables_norm.name == variable_name]
-
-        return dataReceived
+    def callPostTagInsert(self, variables: str | list[str]) -> pd.DataFrame | None:
+        """
+        Check if provided tag names exist, then create them if not
+        Args:
+            variables (str or list): variable name or names to be created
+        Returns:
+            response: Dataframe of Nexus info for tags created
+            """
+        # Crear variables que no existen:
+        # 1. Si se ha proporcionado un str, pasar a lista
+        if isinstance(variables, str):
+            variables = [variables]
+        body = json.dumps(variables)
+        self.log_print(f'Tags to be created: {body}')
+        url_post = self.url_NX + "/api/Tags/Insert"
+        self.log_print(url_post)
+        return self.__postResponse(url_post, body)
 
     def get_rt_data_tagview(self, uid_tagview: str = None, filters=None):
         """
@@ -322,6 +415,33 @@ class APINexus:
 
         return dataReceived
 
+    def callPostValue_operate_mult(self, df: pd.DataFrame):
+        """Escritura de variable en tiempo real en PLC.
+        Args:
+            df: dataframe con los campos ['uid', 'value'] de las variables a escribir
+        """
+
+        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
+        url_completa = self.url_NX + "/api/Tags/operate"
+        df = df[['uid', 'value']]
+        payload = df.to_json(orient='records')
+
+        headers = {
+            "nexustoken": self.token,
+            "nexusapiversion": self.version,
+            "Content-Type": "application/json",
+        }
+        response = requests.request(
+            "POST", url_completa, headers=headers, data=payload
+        )
+
+        if response.status_code == 200:
+            dataReceived = "Escritura correcta"
+        else:
+            dataReceived = response.content
+        self.log_print(response.text.encode("utf8"))
+        return dataReceived
+
     def callPostValue_operate(self, variable_uid, variable_value):
         """Escritura de variable en tiempo real.
         args:
@@ -540,16 +660,16 @@ class APINexus:
 
         return dataReceived
 
-    def filter_installation(self, Datefrom, Dateto, columnas, filter_txt=None, uids=None, resolucion=3,
+    def filter_installation(self, datefrom: datetime.datetime, dateto: datetime.datetime, columnas=None,
+                            filter_txt=None, uids=None, resolucion=3,
                             fuente=0) -> pd.DataFrame:
         """
         la funcion recibe como parametros la fecha ini, fecha fin, un df con los uid y
         los nombres de las variables de la instalación y el filtro de texto aplicar
-
         Args:
-            Datefrom (datetime)
-            Dateto (datetime)
-            columnas (dataframe)
+            datefrom (datetime)
+            dateto (datetime)
+            columnas (dataframe): dataframe con la información de los tags a leer. Debe contener menos de 100 tags para versiones mayores a v3. Si no se proporciona, usa el de toda la instalacion
             uids: (optional) (list): lista de los uids de las variables deseadas. No puede usarse con filter_txt
             filter_txt (optional: list or str): lista de nombres de las variables deseadas. No puede usarse con uids
             resolucion (int): de 0 a 10 [ RES_30_SEC, RES_1_MIN, RES_15_MIN, RES_1_HOUR, RES_1_DAY, RES_1_MONTH, RES_1_YEAR, RES_5_MIN, RES_200_MIL, RES_500_MIL, RES_1_SEC ]
@@ -557,9 +677,12 @@ class APINexus:
         Returns:
             filtered_hist (dataframe)
         """
-        # Check parameters
         if uids is not None and filter_txt is not None:
             raise ValueError('Error en uids y filter_txt. No pueden usarse ambos parámetros de búsqueda')
+        # Check parameters
+        if columnas is None:
+            datos = self.callGetTags()
+            columnas = json_normalize(datos)
         if uids is None:
             uids = []
             if isinstance(filter_txt, list):
@@ -570,14 +693,39 @@ class APINexus:
                 if filter_txt:
                     uids = list(columnas[columnas['name'].str.contains(filter_txt, case=False)].uid)
             if not uids:
-                self.log_print('Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
+                self.log_print(
+                    'Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
                 uids = list(columnas.uid)
         # Remove duplicate UIDS
         uids = list(set(uids))
-        fecha_ini = Datefrom
-        fecha_fin = Dateto
-        nexusRequest = NexusRequest(uids, fecha_ini, fecha_fin, fuente, resolucion)
-        filtered_hist = self.callGetTagsHistory(nexusRequest)
+        fecha_ini = datefrom
+        fecha_fin = dateto
+        if self._version_number < 3:
+            nexusRequest = NexusRequest(uids, fecha_ini, fecha_fin, fuente, resolucion)
+            filtered_hist = self.callGetTagsHistory(nexusRequest)
+        elif self._version_number >= 3:
+            # Check that len is less than 100 tags
+            n = len(uids)
+            if n > 100:
+                data = []
+                iter = round(n / 100)
+                for j in range(iter):
+                    ini = j * 100
+                    fin = (j + 1) * 100
+                    if fin < n:
+                        uids_cache = uids[ini:fin]
+                    else:
+                        uids_cache = uids[ini:]
+                    cache = self.callGetTags_History(uids_cache, datefrom.timestamp(), dateto.timestamp(), fuente,
+                                                     resolucion)
+                    data.extend(cache)
+                filtered_hist = data
+
+            else:
+                filtered_hist = self.callGetTags_History(uids, datefrom.timestamp(), dateto.timestamp(), fuente,
+                                                         resolucion)
+        else:
+            raise AttributeError(f'Version provided not implemented {self.version}')
         filtered_hist = self.__json_normalize_check(filtered_hist)
         # Check that there is no error in the dataframe
         self.__check_response_error_df(filtered_hist)
@@ -586,15 +734,16 @@ class APINexus:
         filtered_hist['name'] = filtered_hist['uid'].map(diccio)
         return filtered_hist
 
-    def filter_tagview(self, Datefrom, Dateto, columnas, uid_vista, filter_txt=None, uids=None, resolucion=3,
+    def filter_tagview(self, datefrom: datetime.datetime, dateto: datetime.datetime, columnas: pd.DataFrame,
+                       uid_vista: str, filter_txt=None, uids=None, resolucion=3,
                        fuente=0) -> pd.DataFrame:
         """
         La funcion recibe como parametros la fecha ini, fecha fin, un df con los uid y
         los nombres de las variables de la instalación y el filtro de texto aplicar [pueden ser varios]
         Args:
-            Datefrom (datetime)
-            Dateto (datetime)
-            columnas (dataframe)
+            datefrom (datetime)
+            dateto (datetime)
+            columnas (dataframe): Dataframe con la informacion de las variables en la vista devuelto por callGetTagViews
             uid_vista: uid de la vista en la que se busca
             uids: (optional) (list): lista de los uids de las variables deseadas. No puede usarse con filter_txt
             filter_txt (optional: list or str): lista de nombres de las variables deseadas. No puede usarse con uids
@@ -616,17 +765,21 @@ class APINexus:
                 if filter_txt:
                     uids = list(columnas[columnas['name'].str.contains(filter_txt, case=False)].uid)
             if not uids:
-                self.log_print('Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
+                self.log_print(
+                    'Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
                 uids = list(columnas.uid)
         # Remove duplicate UIDS
         uids = list(set(uids))
-        fecha_ini = Datefrom
-        fecha_fin = Dateto
-        # fuente: [0 -->RAW, 1 -->STATS_PER_HOUR, 2 -->STATS_PER_DAY, 3 -->STATS_PER_MONTH, 4 -->TRANSIENT]
-        # resolucion: de 0 a 10 [ RES_30_SEC, RES_1_MIN, RES_15_MIN, RES_1_HOUR, RES_1_DAY, RES_1_MONTH, RES_1_YEAR,
-        # RES_5_MIN, RES_200_MIL, RES_500_MIL, RES_1_SEC ]
-        nexusRequest = NexusRequest(uids, fecha_ini, fecha_fin, fuente, resolucion)
-        filtered_hist = self.callGetDataviewHistory(uid_vista, nexusRequest)
+        fecha_ini = datefrom
+        fecha_fin = dateto
+        if self._version_number < 3:
+            nexusRequest = NexusRequest(uids, fecha_ini, fecha_fin, fuente, resolucion)
+            filtered_hist = self.callGetDataviewHistory(uid_vista, nexusRequest)
+        elif self._version_number >= 3:
+            filtered_hist = self.callGetDataview_History(uid_vista, uids, datefrom.timestamp(), dateto.timestamp(),
+                                                         fuente, resolucion)
+        else:
+            raise AttributeError(f'Version provided not implemented {self.version}')
         filtered_hist = self.__json_normalize_check(filtered_hist)
         # Check dataframe for errors
         self.__check_response_error_df(filtered_hist)
@@ -646,10 +799,11 @@ class APINexus:
         groups = [groups] if isinstance(groups, str) else groups
         uids = []
         for group in groups:
-            uids.extend(alarms['uid'][alarms['group'].str.contains(group, case=False)].to_list())
+            uids.extend(alarms['uid'][alarms['groupName'].str.contains(group, case=False)].to_list())
         return uids
 
-    def __check_response_error_df(self, filtered_hist: pd.DataFrame) -> None:
+    @staticmethod
+    def __check_response_error_df(filtered_hist: pd.DataFrame) -> None:
         """
         Checks the received dataframe from API request and raises an error if not valid
         """
